<?php
/**
 * @file
 * Drupal Inspect module profiler class
 */

/**
 * Execution time profiler.
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
class InspectProfile extends InspectBs {
  /**
   * Instance to enable filing stats at object destruction.
   *
   * Boolean FALSE, if user not allowed to profile.
   * 
   * @type InspectProfile|FALSE|NULL
   */
  protected static $_o;

  /**
   * @return InspectProfile
   */
  protected function  __construct() {}

  /**
   * Files profile statistics.
   *
   * Files to directory private://inspect/profiles.
   *
   * To limit the number of files, and still avoid concurrent writing, appends to file named [date]_user_[user id];
   * plus, for anonymous user only, _[inspect session number].
   * 
   * @return void
   */
  public function __destruct() {
    if (self::$_profiles) {
      file_put_contents(
          self::$_profileDir . '/profile_'
              . date('Ymd', $tReq = (int) round(self::$_requestTimeMillis / 1000))
              . '_user_' . self::$_userId
              . (self::$_userId ? '' : ('_' . self::$_sessionNo))
              . '.log',
          self::$_sessionNo . ':' . self::$_requestNo
              . ' - ' . $_SERVER['REQUEST_METHOD'] . ' ' . $_SERVER['REQUEST_URI']
              . self::$_options['delimiters'][1]
              . date('Y-m-d H:i:s', $tReq) . ' - user: ' . self::$_userId . ' - '
              . self::_stat() . "\n\n",
          FILE_APPEND
      );
    }
  }

  /**
   * Copy of InspectBs::$_permissions, or FALSE.
   *
   * @var array|bool $_permissions
   */
  protected static $_permissions;

  /**
   * @var integer|string $_userId
   */
  protected static $_userId;

  /**
   * @var string $_profileDir
   */
  protected static $_profileDir;

  /**
   * @var integer $_requestTimeMillis
   */
  protected static $_requestTimeMillis;

  /**
   * @var array $_names
   */
  protected static $_names = array();

  /**
   * @var array $_profiles
   */
  protected static $_profiles = array();

  /**
   * @var array $_intervals
   */
  protected static $_intervals = array();

  /**
   * @var array $_splitTimes
   */
  protected static $_splitTimes = array();

  /**
   * Call this before actual profiling, or initialization may influence measurement.
   *
   * Callable any number of times, only initializes once.
   *
   * Does not call parent InspectBs::_init(), because the basic methods of this class doesnt need that.
   * However ::log() and ::get() call that parent initializer.
   * @return bool
   *  - FALSE if no permission
   */
  public static function init() {
    if (self::$_o) {
      return TRUE;
    }
    if (self::$_o === FALSE) {
      return FALSE;
    }
    if (!array_key_exists('user', $GLOBALS) || !is_object($GLOBALS['user']) || !property_exists($GLOBALS['user'], 'uid')) {
      return FALSE; // Dont set self::$_o to FALSE, later attempt may succeed in this respect.
    }
    if (!user_access('inspect profile')
        || !($le = strlen($dir = variable_get('file_private_path', FALSE)))) {
      return (self::$_o = FALSE);
    }
    //  Establish filing directory (1).
    if ($dir{$le - 1} == '/') { // Remove trailing slash.
      $dir = substr($dir, 0, $le - 1);
      $le -= 1;
    }
    if($le < 2) { // Private files dir cannot be shorter than 2 chars.
      return (self::$_o = FALSE);
    }

    //  This may work, set other singleton properties.
    self::$_userId = $GLOBALS['user']->uid;
    self::$_requestTimeMillis =
          ($_SERVER[ array_key_exists('REQUEST_TIME_FLOAT', $_SERVER) ? 'REQUEST_TIME_FLOAT' : 'REQUEST_TIME']) * 1000;
    self::_initSessionNo();

    //  Establish filing directory (2).
    if (is_dir($profileDir = $dir . '/inspect/profiles')) {
      self::$_profileDir = $profileDir;
      self::$_o = new InspectProfile();
      return TRUE;
    }
    if (file_exists($profileDir)) { // File, not dir.
      return (self::$_o = FALSE);
    }
    //  Cant safely use recursive mkdir, because cant establish file mode for certain.
    $dir .= '/inspect';
    if ( (!file_exists($dir) && !mkdir($dir)) || !is_dir($dir)
        || !mkdir($profileDir)) {
      return (self::$_o = FALSE);
    }
    self::$_profileDir = $profileDir;
    self::$_o = new InspectProfile();
    return TRUE;
  }

  /**
   * Register event.
   *
   * If both args $start and $stop are falsy: simply marks a point in time.
   *
   * Profile statistics will be written to file at shutdown, in the directory private://inspect/profiles.
   * @param string $name
   *  - will be truncated to length 32
   *  - empty not allowed
   * @param boolean $stop
   *  - default: FALSE
   *  - truthy: stop interval and clear starting point
   *  - ignored if no starting point currently registered
   * @param boolean $start
   *  - default: FALSE
   *  - truthy: start new interval
   *  - ignored if a starting point currently is registered
   * @param string $message
   *  - default: empty
   * @return bool|NULL
   *   - NULL if no permission
   *   - FALSE on error
   */
  public static function event($name, $stop = FALSE, $start = FALSE, $message = '') {
    $t = microtime(TRUE) * 1000;
    if (!($le = strlen($nm = '' . $name))) {
      return FALSE;
    }
    if (!self::$_o && (self::$_o === FALSE || !static::init())) {
      return NULL;
    }
    if ($le > 32) {
      $nm = substr($nm, 0, 32);
    }
    self::$_names[$nm] = 1;
    $tSplit = -1;
    if ($start || $stop) {
      $intervals =& self::$_intervals;
      $hasStart = isset($intervals[$nm]);
      //  Stop if currently a start.
      if ($stop && $hasStart) {
        $tSplit = $t - $intervals[$nm];
        if (isset(self::$_splitTimes[$nm])) {
          self::$_splitTimes[$nm][] = $tSplit;
        }
        else {
          self::$_splitTimes[$nm] = array(
              0 => $tSplit
          );
        }
        if ($start) {
          $hasStart = FALSE; // Because stopped now.
        }
        else {
          unset($intervals[$nm]);
        }
      }
      else {
        $stop = FALSE;
      }
      //  Start if currently no start.
      if ($start && !$hasStart) {
        $intervals[$nm] = $t;
      }
      else {
        $start = FALSE;
      }
    }
    self::$_profiles[] = array(
        $nm,
        $t,
        $tSplit,
        0 + ($stop ? 2 : 0) + ($start ? 1 : 0),
        $message,
    );
    return TRUE;
  }

  /**
   * Log profile statistics to database (watchdog), if permitted.
   * @param string $name
   *  - default: empty (~ list all profiles)
   *  - non-empty will be truncated to length 32
   * @param string $message
   *  - default: empty
   * @return bool|NULL
   *   - NULL if never initialized or user isnt permitted to log inspections
   *   - FALSE on error
   */
  public static function log($name = '', $message = '') {
    if (!self::$_o || (!$permissions = self::_init()) || !$permissions['inspect log']) {
      return NULL;
    }
    $opts =& self::$_options;
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag_log'])) {
      $tagStart = '<' . $u . (empty($opts['no_folding']) ? ' class="module-inspect-collapsible inspect-profile"' : '') . '>';
      $tagEnd = '</' . $u . '>';
    }
    $nl = $opts['newline_log'];
    return self::log_db(
        $tagStart
          . ( empty($message) ? '' : ($message . ':' . $nl) )
          . '[Inspect profile - ' . self::$_sessionNo . ':' . self::$_requestNo . ']' . $nl
          . self::_stat($name)
          . $tagEnd,
        'inspect profile',
        WATCHDOG_INFO
    );
  }

  /**
   * Get profile statistics as string, if permitted.
   * @param string $name
   *  - default: empty (~ list all profiles)
   *  - non-empty will be truncated to length 32
   * @param string $message
   *  - default: empty
   * @return string
   *   - empty if never initialized or user isnt permitted to get inspections, or on error
   */
  public static function get($name = '', $message = '') {
    if (!self::$_o || (!$permissions = self::_init()) || !$permissions['inspect get']) {
      return '';
    }
    $opts =& self::$_options;
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag_get'])) {
      $tagStart = '<' . $u . (empty($opts['no_folding']) ? ' class="module-inspect-collapsible inspect-profile"' : '') . '>';
      $tagEnd = '</' . $u . '>';
    }
    $nl = $opts['newline_get'];
    return $tagStart
        . ( empty($message) ? '' : ($message . ':' . $nl) )
        . '[Inspect profile]' . $nl
        . self::_stat($name)
        . $tagEnd;
  }

  /**
   * @param string $name
   *  - default: empty (~ all)
   * @return string
   */
  protected static function _stat($name = '') {
    $ps =& self::$_profiles;
    $delim = ' | ';
    $nl = self::$_options['delimiters'][1];
    $lgnd = 'call ' . $delim;
    $tReq = self::$_requestTimeMillis;
    $aSplits =& self::$_splitTimes;
    $tLast = 0;
    //  Non-empty arg $name.
    if (($filter = ($le = strlen($name = '' . $name)) ? TRUE : FALSE)) {
      if ($le > 32) {
        $name = substr($name, 0, 32);
      }
      $s = 'Profile: ' . $name;
      $av = NULL;
      if(isset($aSplits[$name])) {
        $av = array_sum($aSplits[$name]) / count($aSplits[$name]);
      }
    }
    else { // All profiles.
      $lNms = 0;
      $le = count($names = array_keys(self::$_names));
      for ($i = 0; $i < $le; $i++) {
        if (($lNm = strlen($names[$i])) > $lNms) {
          $lNms = $lNm;
        }
      }
      $lgnd .= str_pad('name', $lNms, ' ') . $delim;
      unset($names);
      $s = 'All profiles';
      $padNames = array();
      $avs = array();
      foreach ($aSplits as $k => &$v) {
        $avs[$k] = array_sum($v) / count($v);
      }
      unset($v);
    }
    $s .= $nl . 'unit: milliseconds' . $nl . $lgnd
        . join($delim, array('in request ', 'since last ', 'split time ', 'vs. average', 'stop', 'start', 'message'));
    if (!($le = count($ps))) {
      return $s . '- none -';
    }
    $j = 0;
    for ($i = 0; $i < $le; $i++) {
      $p = $ps[$i];
      $nm = $p[0];
      if (!$filter) {
        if (array_key_exists($nm, $padNames)) {
          $padNm = $padNames[$nm];
        }
        else {
          $padNames[$nm] = $padNm = str_pad($nm, $lNms, ' ');
        }
      }
      else if ($nm != $name) {
        continue;
      }
      if(($tSplit = $p[2]) > -1) {
        $vsAv_sign = '+';
        if (($vsAv = $tSplit - (!$filter ? $avs[$nm] : $av)) < 0) {
          $vsAv_sign = '-';
          $vsAv *= -1;
        }
      }
      $ss = $delim;
      switch ($p[3]) { // Stop/start.
        case 3:
          $ss .= '  ';
          break;
        case 2:
          $ss .= '  stop    ';
          break;
        case 1:
          $ss .= '  start   ';
          break;
        default:
          $ss .= '          ';
      }
      $t = $p[1]; // Call time.
      $s .= $nl
          . (
              ($n = ($j + 1)) < 10 ? '    ' :
              ($n < 100 ? '   ' :
              ($n < 1000 ? '  ' :
              ($n < 10000 ? ' ' : '')))
            )
          . $n . $delim
          . (!$filter ? ($padNm . $delim) : '')
          . (
              ($n = ($t - $tReq)) < 10 ? '     ' :
              ($n < 100 ? '    ' :
              ($n < 1000 ? '   ' :
              ($n < 10000 ? '  ' :
              ($n < 100000 ? ' ' : ''))))
            )
            . number_format(round($n, 4), 4, '.', '')
          . $delim
          . ($j ? // Compute time since last call (any name, if no $name).
              (
                (($n = ($t - $tLast)) < 10 ? '     ' :
                ($n < 100 ? '    ' :
                ($n < 1000 ? '   ' :
                ($n < 10000 ? '  ' :
                ($n < 100000 ? ' ' : '')))))
                . number_format(round($n, 4), 4, '.', '')
              ) : '     0     '
            )
          . $delim
          . (($n = $tSplit) > -1 ? // Split time.
              (
                ($n < 10 ? '     ' :
                ($n < 100 ? '    ' :
                ($n < 1000 ? '   ' :
                ($n < 10000 ? '  ' :
                ($n < 100000 ? ' ' : '')))))
                . number_format(round($n, 4), 4, '.', '')
              ) : '           '
            )
          . $delim
          . ($tSplit > -1 ? // Then versus average.
              (
                ($vsAv < 10 ? '    ' :
                ($vsAv < 100 ? '   ' :
                ($vsAv < 1000 ? '  ' :
                ($vsAv < 10000 ? ' ' : ''))))
                . $vsAv_sign
                . number_format(round($vsAv, 4), 4, '.', '')
              ) : '           '
            )
          . $delim
          . (($n = $p[3]) > 1 ? // Stop/start.
              ' 1  ' : '    '
            )
          . $delim
          . (($n = $p[3]) % 2 ? // Stop/start.
              '  1  ' : '     '
            )
          . $delim
          . (!strlen($n = $p[4]) ? '' :
              htmlspecialchars(
                  str_replace(array("\r", "\n", "\t", '|'), array('', ' ', ' ', '/'), $n), // Message.
                  ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
                  'UTF-8',
                  FALSE // No double encoding
              )
            );
      $tLast = $t;
      ++$j;
    }
    //  Average split times.
    if (!$filter) {
      $s .= $nl . 'Average split times:';
      foreach ($avs as $nm => $v) {
        $s .= $nl . $nm . ': ' . number_format(round($v, 4), 4, '.', '');
      }
    }
    else if ($av !== NULL) {
      $s .= $nl . 'Average split time: ' . number_format(round($av, 4), 4, '.', '');
    }
    return $s;
  }
}